#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hkvsobekpy', 'hkvsobekpy.core', 'hkvsobekpy.io']

package_data = \
{'': ['*'], 'hkvsobekpy.io': ['.ipynb_checkpoints/*']}

setup(name='hkvsobekpy',
      version='1.4',
      description='package for water-statistics and plausibility checker using his- and bui-file',
      author='HKV',
      author_email='hoek@hkv.nl',
      url='https://www.hkv.nl',
      packages=packages,
      package_data=package_data,
     )
